--[[

Author : Levva of EU Khadgar
	
Version 0.01 - Initial release of beta code to WowAce SVN

Version 0.10 - Working Enh Shammy model, added Blood DK model with HS 
		 upped selections to 10
		 lots of debugging still required
		 
Version 0.20 - Huge refactoring excersise to simply priority code
		Added Hunter model
		Added default priority options - something weird that it doesnt survive relogging though.
		
Version 0.21 - Added Marksman hunter
				
Version 0.30 - Added Elemental Shaman
		Added Blood Deathknight
		Added Shadow Priest
		Added Affliction SB/LT Warlock
		Added Prot Paladin
		Fix priorities being reset to none on logout
		
Version 0.31 - Added Rogue, Mage, Warrior, Druid code - mostly empty at present
		Added Beast Master Hunter
		Fixed bug with Priest broken Mind Blast 
		Some initial Rogue code - needs more work no evenom/eviserate yet for example

--]]

local L = LibStub("AceLocale-3.0"):GetLocale("PriorityQueue")
local AceAddon = LibStub("AceAddon-3.0")
local media = LibStub:GetLibrary("LibSharedMedia-3.0");
PriorityQueue = AceAddon:NewAddon("PriorityQueue", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local REVISION = tonumber(("$Revision: 20 $"):match("%d+"))
local _, _, _, clientVersion = GetBuildInfo()
if clientVersion >= 30100 then
	PriorityQueue.patch31 = true
else
	PriorityQueue.patch31 = false
end
PriorityQueue.class = select(2, UnitClass('player'))
PriorityQueue.constants = {}
local C = PriorityQueue.constants -- Defined in PriorityQueue LUA no locale needed.

PriorityQueue.PriorityFrame = CreateFrame("Frame", "PQ_PriorityFrame", UIParent)
PriorityQueue.msgFrame = CreateFrame("MessageFrame","PriorityQueueMsg",UIParent)

PriorityQueue.textures = {}
PriorityQueue.borders = {}
PriorityQueue.fonts = {}
PriorityQueue.sounds = {}
PriorityQueue.frameBackdrop ={
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		--edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = 12,
		insets = { left = 0, right = 0, top = 0, bottom = 0 }
	}
	
PriorityQueue.priorityTable = {}
PriorityQueue.priorityTable.name = {}
PriorityQueue.priorityTable.icon = {}
PriorityQueue.priorityTable.test = {}
PriorityQueue.priorityTable.name["none"] = L["None"]
PriorityQueue.priorityTable.icon["none"] = nil
PriorityQueue.priorityTable.test["none"] = function () return false end 

PriorityQueue.presets = {}
PriorityQueue.presets.default = "none"
PriorityQueue.presets.priorities = {}
PriorityQueue.presets.priorities["none"] = { "none", "none", "none", "none", "none", "none", "none", "none", "none", "none" }

-------------------
-- Config details
-------------------

PriorityQueue.defaults = {
	char = {
		message = "Welcome Home!",
		relativeTo = "UIParent",
		relativePoint = "CENTER",
		point = "CENTER",
		texture = "Blizzard",
		titlefont = "Friz Quadrata TT",
		titlefontsize = 12,
		titlefonteffect = "none",
		msgfont = "Friz Quadrata TT",
		msgfontsize = 24,
		msgfonteffect = "none",
		xOffset = 0,
		yOffset = -100,
		fWidth = 300,
		fHeight = 175,
		resetOn = true,
		scale = 1,
		debug = false,
		msg = { r = 1, g = .5, b = 0, },
		msgtime = 2,
		newsitem = 0,
		priority = {
			show = true,
			moving = false,
			titleshow = true,
			relativeTo = "UIParent",
			relativePoint = "CENTER",
			point = "CENTER",
			fWidth = 50,
			fHeight = 50,
			xOffset = -185,
			yOffset = -40,
			alpha = 0.4,
			scale = 1,	
			next = 1,
			cooldown = 0.25,
			fstimeleft = 0.5,
			srmana = 5,
			wsmana = 0,
			hwhealth = 25,
			chingroup = true,
			worldbossonly = true,
			combopoints = true,
			usethreatreduction = true,
			name = "none",
			dropthreat = { name = PriorityQueue.priorityTable.name["none"], PriorityQueue.priorityTable.icon["none"] },
			option = { "none", "none", "none", "none", "none", "none", "none", "none", "none", "none" },
		},
		colours = {
			combopoints = { r=1, g= .5,	b=0, a=.8 },
		},
		warning = {
			show = true,
			duration = 3,
			shield = false,
			weapon = true,
			range = false,
			timeleft = 300, 
			colour = { r = 1, g = .5, b = 0, a = 0.5, },
			relativeTo = "UIParent",
			relativePoint = "TOP",
			point = "CENTER",
			fWidth = 400,
			fHeight = 75,
			xOffset = 0,
			yOffset = -250,
		},
	}
}

-----------------------------------------
-- Initialisation & Startup Routines
-----------------------------------------

function PriorityQueue:OnInitialize()
	local AceConfigReg = LibStub("AceConfigRegistry-3.0")
	local AceConfigDialog = LibStub("AceConfigDialog-3.0")

	self.db = LibStub("AceDB-3.0"):New("PriorityQueueDBPC", PriorityQueue.defaults, "char")
	self:DebugPrint("Before option table : current defaults "..self.db.char.priority.name)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("PriorityQueue", self:GetOptions(), {"PriorityQueue", "PQ"} )
	self:DebugPrint("After option table : current defaults "..self.db.char.priority.name)
	media.RegisterCallback(self, "LibSharedMedia_Registered")
	
	-- Add the options to blizzard frame (add them backwards so they show up in the proper order
	self.optionsFrame = AceConfigDialog:AddToBlizOptions("PriorityQueue","PriorityQueue")
	self.db:RegisterDefaults(self.defaults)
	self:DebugPrint("After register defaults : current defaults "..self.db.char.priority.name)
	self:CreatePriorityFrame()
	self:CreateMsgFrame()
	self:News()
	local version = GetAddOnMetadata("PriorityQueue","Version")
	self.version = ("PriorityQueue v%s (r%s)"):format(version, REVISION)
	self:Print(self.version.." Loaded.")
end

function PriorityQueue:OnDisable()
    -- Called when the addon is disabled
end

function PriorityQueue:OnEnable()
	self:LibSharedMedia_Registered()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
--	self:RegisterEvent("CHAT_MSG_SPELL_SELF_DAMAGE")
--	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
--	self:RegisterEvent("UNIT_AURA")
--	self:RegisterEvent("UNIT_MANA")
--	self:RegisterEvent("CHARACTER_POINTS_CHANGED")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
--	self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	self:RegisterEvent("UI_ERROR_MESSAGE")
end

function PriorityQueue:GetDefaultPriorities()
	self:DebugPrint("character class is : "..self.class)
	self.presets.priorities = {}
	self.presets.priorities.name = {}
	self.presets.priorities.list = {}
	self.presets.priorities.name["none"] = L["None"]
	self.presets.priorities.list["none"] = { "none", "none", "none", "none", "none", "none", "none", "none", "none", "none" }
	self:DebugPrint("Before class priorities : current defaults "..self.db.char.priority.name)
	if (self.class == "DEATHKNIGHT") then
		self:GetDKPriorityDefaults()
	elseif (self.class == "DRUID") then
		self:GetDruidPriorityDefaults()
	elseif (self.class == "HUNTER") then
		self:GetHunterPriorityDefaults()
	elseif (self.class == "MAGE") then
		self:GetMagePriorityDefaults()
	elseif (self.class == "PRIEST") then
		self:GetPriestPriorityDefaults()
	elseif (self.class == "PALADIN") then
		self:GetPaladinPriorityDefaults()
	elseif (self.class == "ROGUE") then
		self:GetRoguePriorityDefaults()
	elseif (self.class == "SHAMAN") then
		self:GetShamanPriorityDefaults()
	elseif (self.class == "WARLOCK") then
		self:GetWarlockPriorityDefaults()
	elseif (self.class == "WARRIOR") then
		self:GetWarriorPriorityDefaults()
	end	
	self:DebugPrint("After class priorities : current defaults "..self.db.char.priority.name)
end

function PriorityQueue:LibSharedMedia_Registered()
	for k, v in pairs(media:List("statusbar")) do
		self.textures[v] = v
	end
	for k, v in pairs(media:List("border")) do
		self.borders[v] = v
	end
	for k, v in pairs(media:List("font")) do
		self.fonts[v] = v
	end
	for k, v in pairs(media:List("sound")) do
		self.sounds[v] = v
	end
end

----------------------
-- Event Routines
----------------------

function PriorityQueue:PLAYER_ENTERING_WORLD()
--	self:SetBorderTexture(nil, self.db.char.border)
--	self:RedrawFrames()
	if (not self.db.char.priority.option) then
		self:SetPresetPriority(nil, self.db.char.priority.name)
	end
end

function PriorityQueue:UI_ERROR_MESSAGE()
	if self.db.char.warning.range then
		if arg1 then self:DebugPrint("error message is :"..arg1) end
		if strfind(arg1, L["Out of range"]) or strfind(arg1, L["too far away"]) then
			self:PrintMsg(arg1, self.db.char.warning.colour, 0.5)
		end
	end
end

function PriorityQueue:CHARACTER_POINTS_CHANGED()
	-- probably need to check if player has valid talents for certain priority abilities
end

function PriorityQueue:PLAYER_REGEN_DISABLED() -- Entering Combat
	if self.db.char.priority.show then
		self.db.char.priority.next = "none"
		self:SetPriorityIcon(self.db.char.priority.next)
		self.updatePriority = true
		self.PriorityFrame:Show()
	end
end

function PriorityQueue:PLAYER_REGEN_ENABLED() -- Leaving Combat
	self:SetPriorityIcon("none")
	self.updatePriority = false
	self.PriorityFrame:Hide()
end

function PriorityQueue:PLAYER_TARGET_CHANGED()
	if self.db.char.priority.show and InCombatLockdown() then
		self:SetNextPriority()
	end
end

function PriorityQueue:SPELL_UPDATE_COOLDOWN()
	if self.db.char.priority.show and InCombatLockdown() then
		self:SetNextPriority()
	end
end

function PriorityQueue:UNIT_SPELLCAST_SUCCEEDED()
	if self.db.char.priority.show and InCombatLockdown() then
		self:SetNextPriority()
	end
end

-------------------------------
-- Priority Frame functions
-------------------------------

function PriorityQueue:SetPriorityBackdrop(icon)
	self.PriorityFrame:SetBackdrop({ bgFile = icon,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = 12,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})
end

function PriorityQueue:CreatePriorityFrame()
	self.updatePriority = false
	self.PriorityFrame:SetScale(self.db.char.priority.scale)
	self.PriorityFrame:SetFrameStrata("BACKGROUND")
	self.PriorityFrame:SetWidth(self.db.char.priority.fWidth)
	self.PriorityFrame:SetHeight(self.db.char.priority.fHeight)
	self:SetPriorityBackdrop("Interface/Tooltips/UI-Tooltip-Background")
	self.PriorityFrame:SetBackdropColor(0, 0, 0, self.db.char.priority.alpha);
	self.PriorityFrame:SetMovable(true);
	self.PriorityFrame:RegisterForDrag("LeftButton");
	self.PriorityFrame:SetPoint(self.db.char.priority.point, self.db.char.priority.relativeTo, self.db.char.priority.relativePoint, self.db.char.priority.xOffset, self.db.char.priority.yOffset)
	self.PriorityFrame:SetScript("OnDragStart", 
		function()
			self.PriorityFrame:StartMoving();
		end );
	self.PriorityFrame:SetScript("OnDragStop",
		function()
			self.PriorityFrame:StopMovingOrSizing();
			self.PriorityFrame:SetScript("OnUpdate", nil);
			self:FinishedMoving(self.db.char.priority, self.PriorityFrame);
		end );
	self.PriorityFrame:SetScript("OnUpdate", 
		function()
			if self.updatePriority then
				PriorityQueue.db.char.priority.previous = PriorityQueue.db.char.priority.next
				PriorityQueue:SetNextPriority()
				if PriorityQueue.db.char.priority.next ~= PriorityQueue.db.char.priority.previous then
					PriorityQueue:SetPriorityIcon(PriorityQueue.db.char.priority.next)
				end
			end
		end );
		
	if not self.PriorityFrame.topText then
		self.PriorityFrame.topText = self.PriorityFrame:CreateFontString(nil, "OVERLAY")
	end
	self.PriorityFrame.topText:SetTextColor(1,1,1,1)
	self.PriorityFrame.topText:SetFont(media:Fetch("font", self.db.char.titlefont), self.db.char.titlefontsize)
	self.PriorityFrame.topText:SetPoint("TOP", self.PriorityFrame, "TOP", 0, self.db.char.titlefontsize + 2)
	self.PriorityFrame.topText:SetText(L["Next Priority"])
	self:CreateComboPointFrames()
	
	if self.db.char.priority.show and InCombatLockdown() then
		self.PriorityFrame:Show()
	else
		self.PriorityFrame:Hide()
	end
	if self.db.char.priority.titleshow then
		self.PriorityFrame.topText:Show()
	else
		self.PriorityFrame.topText:Hide()
	end
end

function PriorityQueue:ResetPriority()
	self.PriorityFrame:ClearAllPoints()
	self.db.char.priority.point = self.defaults.char.priority.point
	self.db.char.priority.relativeTo = self.defaults.char.priority.relativeTo 
	self.db.char.priority.relativePoint = self.defaults.char.priority.relativePoint
	self.db.char.priority.xOffset = self.defaults.char.priority.xOffset
	self.db.char.priority.yOffset = self.defaults.char.priority.yOffset
	self.db.char.priority.fWidth = self.defaults.char.priority.fWidth
	self.db.char.priority.fHeight = self.defaults.char.priority.fHeight
	self.db.char.priority.scale = self.defaults.char.priority.scale
	self.PriorityFrame:SetPoint(self.db.char.priority.point, self.db.char.priority.relativeTo, self.db.char.priority.relativePoint, self.db.char.priority.xOffset, self.db.char.priority.yOffset)
	self:CreatePriorityFrame()
	self:SetPresetPriority(nil, PriorityQueue.presets.default)
	self:Print(L["priority_reset"])
end

function PriorityQueue:SetPriorityIcon(priority)
	self:DebugPrint("setting icon to "..priority)
	if InCombatLockdown() or priority == "none" then
		local icon = self.priorityTable.icon[priority]
		self:SetPriorityBackdrop(icon)
	end
end

function PriorityQueue:SetNextPriority()
	if self.db.char.priority.combopoints then
		self:ShowComboPoints()
	end
	if self.priorityTable.test[self.db.char.priority.option[1]]() then
		self.db.char.priority.next = self.db.char.priority.option[1]
	elseif self.priorityTable.test[self.db.char.priority.option[2]]() then
		self.db.char.priority.next = self.db.char.priority.option[2]
	elseif self.priorityTable.test[self.db.char.priority.option[3]]() then
		self.db.char.priority.next = self.db.char.priority.option[3]
	elseif self.priorityTable.test[self.db.char.priority.option[4]]() then
		self.db.char.priority.next = self.db.char.priority.option[4]
	elseif self.priorityTable.test[self.db.char.priority.option[5]]() then
		self.db.char.priority.next = self.db.char.priority.option[5]
	elseif self.priorityTable.test[self.db.char.priority.option[6]]() then
		self.db.char.priority.next = self.db.char.priority.option[6]
	elseif self.priorityTable.test[self.db.char.priority.option[7]]() then
		self.db.char.priority.next = self.db.char.priority.option[7]
	elseif self.priorityTable.test[self.db.char.priority.option[8]]() then
		self.db.char.priority.next = self.db.char.priority.option[8]
	elseif self.priorityTable.test[self.db.char.priority.option[9]]() then
		self.db.char.priority.next = self.db.char.priority.option[9]
	elseif self.priorityTable.test[self.db.char.priority.option[10]]() then
		self.db.char.priority.next = self.db.char.priority.option[10]
	else
		self.db.char.priority.next = "none"
	end	
	if self:CheckThreat() then
		self.db.char.priority.next = self.db.char.priority.dropthreat.name
	end
end

function PriorityQueue:GetDebuffInfo(debuff)
	local index = 1
	while UnitDebuff("target", index) do
		local name, _, _, count, _, _, debuffExpires, isMine = UnitDebuff("target", index)
		if name == debuff and isMine then -- don't return true on finding another player's debuff
			local duration = debuffExpires - GetTime()
			return duration > PriorityQueue.db.char.priority.cooldown, duration
		end
		index = index + 1
	end
	return false, 0
end

function PriorityQueue:GetBuffInfo(buff)
	local index = 1
	while UnitBuff("player", index) do
		local name, _, _, count, _, _, debuffExpires = UnitBuff("player", index)
		if name == debuff then 
			local duration = debuffExpires - GetTime()
			return duration > PriorityQueue.db.char.priority.cooldown, duration
		end
		index = index + 1
	end
	return false, 0 
end

function PriorityQueue:SpellAvailable(spellname)
	local start, duration = GetSpellCooldown(spellname)
	if start then
		local timeleft = start + duration - GetTime()
		return timeleft <= PriorityQueue.db.char.priority.cooldown, duration
	end
	return false, 999
end

function PriorityQueue:GetExpectedGCD()
	local gcd = self:GetExpectedCastTime(1.5)
	if gcd < 1 then 
		return 1
	else
		return gcd
	end
end

function PriorityQueue:GetExpectedCastTime(baseCastTime)
	local hasteRating = GetCombatRatingBonus(CR_HASTE_SPELL) / 100
	return baseCastTime/(1 + hasteRating)
end

function PriorityQueue:CheckThreat()
	if not self.db.char.priority.dropthreat.name == "none" and self.db.priority.usethreatreduction then
		if not self.db.char.priority.worldbossonly or (self.db.char.priority.worldbossonly and UnitClassification("target") == "worldboss") then
			if (GetNumPartyMembers()>=4 or GetNumRaidMembers()>=4) then
				-- don't use threat reduction option if soloing
				local _, state, _, raw = UnitDetailedThreatSituation("player", "target")
				-- state 2 or 3 = tanking which is not what a dps should be doing
				-- state 0 threat < 100%, state 1 threat >=100%
				if not state then state = 0 end
				if not raw then raw = 0 end
				if  state > 1 or raw > self.db.char.threatThreshold then
					-- we have exceeded threshold or have already started tanking so threat reduction needed.
					local startTime, duration = GetSpellCooldown(self.db.char.priority.dropthreat.name)
					if startTime then -- extra check in case player hasn't learned spell
						local timeleft = startTime + duration - GetTime()
						if timeleft <= PriorityQueue.db.char.priority.cooldown then -- check its off cooldown
							return true
						end
					end
				end
			end
		end
	end
	return false
end

--------------------------
-- Combo Points Frame
--------------------------

function PriorityQueue:ShowComboPoints()
--~ 	local col = self.db.char.colours.maelstrom
--~ 	for index = 1, 5 do
--~ 		if self.PriorityFrame.combo[index].frame then
--~ 			if self.db.char.msstacks >= index then
--~ 				self.PriorityFrame.combo[index].frame:SetBackdropColor(1, 0, 0, 1)
--~ 			else
--~ 				self.PriorityFrame.combo[index].frame:SetBackdropColor(col.r, col.g, col.b, 0)
--~ 			end
--~ 		else
--~ 			self:Print("Error could not find combo frame "..index)
--~ 		end
--~ 	end
end

function PriorityQueue:CreateComboPointFrames()
	self.PriorityFrame.combo = {}
	for index = 1, 5 do 
		self:CreateComboPoint(index)
	end
end
	
function PriorityQueue:CreateComboPoint(index)
	if not self.PriorityFrame.combo[index] or not self.PriorityFrame.combo[index].frame then
		self.PriorityFrame.combo[index] = {}
		self.PriorityFrame.combo[index].frame = CreateFrame("Frame", "SAA_PriorityComboFrame"..index, self.PriorityFrame)
	end
	local comboFrame = self.PriorityFrame.combo[index].frame
	local width = self.db.char.priority.fWidth / 5
	local height = self.db.char.priority.fHeight / 5
	comboFrame:SetScale(self.db.char.priority.scale)
	comboFrame:SetFrameStrata("BACKGROUND")
	comboFrame:SetWidth(width)
	comboFrame:SetHeight(height)
	comboFrame:SetBackdrop({ bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = 4,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})	
	local col = self.db.char.colours.combopoints
	comboFrame:SetBackdropColor(0, 0, 0, 0);
	comboFrame:SetPoint("BOTTOMLEFT",  self.PriorityFrame, "BOTTOMLEFT", (index - 1) * width , - height)

	if self.db.char.priority.combopoints then
		comboFrame:Show()
	else
		comboFrame:Hide()
	end
end

---------------------------
-- functions
---------------------------

function PriorityQueue:DisplayVersion()
	self:Print(self.version)
end

function PriorityQueue:GetGCD(spellname)
	self:DebugPrint("check gcd on "..spellname)
	local startTime, duration, enabled = GetSpellCooldown(spellname) 
	duration = duration or 0 -- force nil to 0 if spell isnt available.
	if duration > 0 then 
		return startTime + duration - GetTime()
	end
	return 0
end

---------------------------
-- Message Frame
---------------------------

function PriorityQueue:CreateMsgFrame()
	self.msgFrame:ClearAllPoints()
	self.msgFrame:SetWidth(400)
	self.msgFrame:SetHeight(75)
	self.msgFrame:SetFrameStrata("BACKGROUND")
	self.msgFrame:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		--edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = 12,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})
	self.msgFrame:SetBackdropColor(1, 1, 1, 0)
	self.msgFrame:SetMovable(true)
	self.msgFrame:RegisterForDrag("LeftButton")
	self.msgFrame:SetScript("OnDragStart", 
		function()
			self.msgFrame:StartMoving();
		end );
	self.msgFrame:SetScript("OnDragStop",
		function()
			self.msgFrame:StopMovingOrSizing();
			self.msgFrame:SetScript("OnUpdate", nil);
			self:FinishedMoving(self.db.char.warning, self.msgFrame);
		end );
	self.msgFrame:SetPoint(self.db.char.warning.point, self.db.char.warning.relativeTo, self.db.char.warning.relativePoint, self.db.char.warning.xOffset, self.db.char.warning.yOffset)
	self.msgFrame:SetInsertMode("TOP")
	self.msgFrame:SetFrameStrata("HIGH")
	self.msgFrame:SetToplevel(true)
	local font = media:Fetch("font", self.db.char.msgfont)
	self.msgFrame:SetFont(font, self.db.char.msgfontsize, self.db.char.msgfonteffect)
	self.msgFrame:Show()
end

function PriorityQueue:PrintMsg(msg, col, time)
	if self.db.char.warning.show then
		if col == nil then
			col = { r=1, b=1, g=1, a=1 }
		end
		if time == nil then 
			time = 3
		end
		if time ~= 5 then
			self.msgFrame:SetTimeVisible(time)
		end
		self.msgFrame:AddMessage(msg, col.r, col.g, col.b, 1, col.time)
	end
end

function PriorityQueue:DebugPrint(msg)
	if self.db.char.debug then
		self:Print(msg)
	end
end