if not PriorityQueue then return end

if select(2, UnitClass('player')) ~= "WARLOCK" then
	return
end

local L = LibStub("AceLocale-3.0"):GetLocale("PriorityQueue")
local C = PriorityQueue.constants -- Defined in PriorityQueue LUA no locale needed.

-- I hate typing a lot...
local function spell(name, id)
	C[name], _, C[name .. " Icon"], C[name .. " Mana"], _, _, C[name .. " Cast Time"] = GetSpellInfo(id)
end

spell("Haunt", 59164)
spell("Unstable Affliction", 47843)
spell("Immolate", 47811)
spell("Curse of Agony", 47864)
spell("Corruption", 47813)
spell("Siphon Life", 47862)
spell("Shadow Bolt", 47809)
spell("Soulshatter", 29858)
spell("Life Tap", 57946)

function PriorityQueue:GetWarlockPriorityDefaults()
	self.presets.priorities.name["Affliction SB/LT"] =  L["Affliction SB/LT"]
	-- self.presets.priorities.name["Affliction SB/DP"] =  L["Affliction SB/DP"]
	-- self.presets.priorities.name["Affliction Fire/DP"] =  L["Affliction Fire/DP"]
	-- self.presets.priorities.name["Demon SB"] =  L["Demon SB"]
	-- self.presets.priorities.name["Demon Fire"] =  L["Demon Fire"]
	-- self.presets.priorities.name["Destruction Fire"] =  L["Destruction Fire"]
	self.presets.priorities.list["Affliction SB/LT"] = {"lt", "h", "ua", "i", "coa", "c", "sl", "sb", "none", "none" }
	self.presets.default = "Affliction SB/LT"

	self.db.char.priority.dropthreat.name = C["Soulshatter"]
	self.db.char.priority.dropthreat.icon = C["Soulshatter Icon"]

	non_stacking_debuff("h", "Haunt")
	non_stacking_debuff("ua", "Unstable Affliction")
	non_stacking_debuff("i", "Immolate")
	non_stacking_debuff("coa", "Curse of Agony")
	non_stacking_debuff("c", "Corruption")
	non_stacking_debuff("sl", "Siphon Life")
	destruction("sb", "Shadow Bolt")
	lifetap("lt", "Life Tap")
end

local function non_stacking_debuff(x, y)
	PriorityQueue.priorityTable.name[x] = C[y]
	PriorityQueue.priorityTable.icon[x] = C[y .. " Icon"]
	local needed = C[y .. " Mana"]
	local casttime = C[y .. " Cast Time"] / 1000
	local endtime = 0
	-- each spell has a different flight speed, and the time depends on
	-- the distance to the target - we just assume half a second for
	-- all our dots
	local flighttime = 0.5
	PriorityQueue.priorityTable.test[x] = function () 
		local casting, _, _, _, _, t = UnitCastingInfo("player")
		-- It doesn't stack, so never cast if it's already casting.
		if casting == C[y] then
			endtime = t / 1000
			return false
		end
		if GetTime() - endtime < flighttime then
			return false
		end
		local mana = UnitPower("player")
		-- Instead of 1000, we should use the mana of the currently
		-- being cast spell...
		if mana < 1000 + needed then return false end
		local dot, duration = PriorityQueue:GetDebuffInfo(C[y])
		if dot then
			if duration > PriorityQueue.db.char.priority.cooldown + casttime then
				return false
			end
		end
		return PriorityQueue:SpellAvailable(C[y])
	end 
end

local function destruction(x, y)
	PriorityQueue.priorityTable.name[x] = C[y]
	PriorityQueue.priorityTable.icon[x] = C[y .. " Icon"]
	local needed = C[y .. " Mana"]
	PriorityQueue.priorityTable.test[x] = function ()
		local mana = UnitPower("player")
		if mana < needed then return false end
		return PriorityQueue:SpellAvailable(C[y])
	end
end

local function lifetap(x, y)
	PriorityQueue.priorityTable.name[x] = C[y]
	PriorityQueue.priorityTable.icon[x] = C[y .. " Icon"]
	PriorityQueue.priorityTable.test[x] = function () 
		local mana = UnitPower("player")
		local ava = PriorityQueue:SpellAvailable(C[y])
		-- If we have more mana and come here something went wrong.
		return ava and mana < 1000
	end 
end

